#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy

from openstackclient.identity.v3 import implied_role
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestRole(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        identity_client = self.identity_client

        # Get a shortcut to the UserManager Mock
        self.users_mock = identity_client.users
        self.users_mock.reset_mock()

        # Get a shortcut to the UserManager Mock
        self.groups_mock = identity_client.groups
        self.groups_mock.reset_mock()

        # Get a shortcut to the DomainManager Mock
        self.domains_mock = identity_client.domains
        self.domains_mock.reset_mock()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = identity_client.projects
        self.projects_mock.reset_mock()

        # Get a shortcut to the RoleManager Mock
        self.roles_mock = identity_client.roles
        self.roles_mock.reset_mock()

        # Get a shortcut to the InferenceRuleManager Mock
        self.inference_rules_mock = identity_client.inference_rules
        self.inference_rules_mock.reset_mock()

    def _is_inheritance_testcase(self):
        return False


class TestImpliedRoleCreate(TestRole):
    def setUp(self):
        super().setUp()

        self.roles_mock.list.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLES[0]),
                loaded=True,
            ),
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLES[1]),
                loaded=True,
            ),
        ]

        fake_resource = fakes.FakeResource(
            None,
            {
                'prior_role': copy.deepcopy(identity_fakes.ROLES[0]),
                'implied': copy.deepcopy(identity_fakes.ROLES[1]),
            },
            loaded=True,
        )
        self.inference_rules_mock.create.return_value = fake_resource

        self.cmd = implied_role.CreateImpliedRole(self.app, None)

    def test_implied_role_create(self):
        arglist = [
            identity_fakes.ROLES[0]['id'],
            '--implied-role',
            identity_fakes.ROLES[1]['id'],
        ]
        verifylist = [
            ('role', identity_fakes.ROLES[0]['id']),
            ('implied_role', identity_fakes.ROLES[1]['id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # InferenceRuleManager.create(prior, implied)
        self.inference_rules_mock.create.assert_called_with(
            identity_fakes.ROLES[0]['id'], identity_fakes.ROLES[1]['id']
        )

        collist = ('implied', 'prior_role')
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.ROLES[1]['id'],
            identity_fakes.ROLES[0]['id'],
        )
        self.assertEqual(datalist, data)


class TestImpliedRoleDelete(TestRole):
    def setUp(self):
        super().setUp()

        self.roles_mock.list.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLES[0]),
                loaded=True,
            ),
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLES[1]),
                loaded=True,
            ),
        ]

        fake_resource = fakes.FakeResource(
            None,
            {
                'prior-role': copy.deepcopy(identity_fakes.ROLES[0]),
                'implied': copy.deepcopy(identity_fakes.ROLES[1]),
            },
            loaded=True,
        )
        self.inference_rules_mock.delete.return_value = fake_resource

        self.cmd = implied_role.DeleteImpliedRole(self.app, None)

    def test_implied_role_delete(self):
        arglist = [
            identity_fakes.ROLES[0]['id'],
            '--implied-role',
            identity_fakes.ROLES[1]['id'],
        ]
        verifylist = [
            ('role', identity_fakes.ROLES[0]['id']),
            ('implied_role', identity_fakes.ROLES[1]['id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.inference_rules_mock.delete.assert_called_with(
            identity_fakes.ROLES[0]['id'], identity_fakes.ROLES[1]['id']
        )


class TestImpliedRoleList(TestRole):
    def setUp(self):
        super().setUp()

        self.inference_rules_mock.list_inference_roles.return_value = (
            identity_fakes.FakeImpliedRoleResponse.create_list()
        )

        self.cmd = implied_role.ListImpliedRole(self.app, None)

    def test_implied_role_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.inference_rules_mock.list_inference_roles.assert_called_with()

        collist = [
            'Prior Role ID',
            'Prior Role Name',
            'Implied Role ID',
            'Implied Role Name',
        ]
        self.assertEqual(collist, columns)
        datalist = [
            (
                identity_fakes.ROLES[0]['id'],
                identity_fakes.ROLES[0]['name'],
                identity_fakes.ROLES[1]['id'],
                identity_fakes.ROLES[1]['name'],
            )
        ]
        x = [d for d in data]
        self.assertEqual(datalist, x)



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from openstackclient.network.v2 import network_auto_allocated_topology
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit.network.v2 import fakes as network_fakes


class TestAutoAllocatedTopology(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()
        self.projects_mock = self.identity_client.projects


class TestCreateAutoAllocatedTopology(TestAutoAllocatedTopology):
    project = identity_fakes.FakeProject.create_one_project()
    network_object = network_fakes.create_one_network()

    topology = network_fakes.create_one_topology(
        attrs={'id': network_object.id, 'project_id': project.id}
    )

    columns = (
        'id',
        'project_id',
    )

    data = (
        network_object.id,
        project.id,
    )

    def setUp(self):
        super().setUp()

        self.cmd = network_auto_allocated_topology.CreateAutoAllocatedTopology(
            self.app, None
        )
        self.network_client.get_auto_allocated_topology.return_value = (
            self.topology
        )

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.get_auto_allocated_topology.assert_called_with(
            None
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_project_option(self):
        arglist = [
            '--project',
            self.project.id,
        ]

        verifylist = [
            ('project', self.project.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.get_auto_allocated_topology.assert_called_with(
            self.project.id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_project_domain_option(self):
        arglist = [
            '--project',
            self.project.id,
            '--project-domain',
            self.project.domain_id,
        ]

        verifylist = [
            ('project', self.project.id),
            ('project_domain', self.project.domain_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.get_auto_allocated_topology.assert_called_with(
            self.project.id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_or_show_option(self):
        arglist = [
            '--or-show',
        ]

        verifylist = [
            ('or_show', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.get_auto_allocated_topology.assert_called_with(
            None
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestValidateAutoAllocatedTopology(TestAutoAllocatedTopology):
    project = identity_fakes.FakeProject.create_one_project()
    network_object = network_fakes.create_one_network()

    topology = network_fakes.create_one_topology(
        attrs={'id': network_object.id, 'project_id': project.id}
    )

    columns = (
        'id',
        'project_id',
    )

    data = (
        network_object.id,
        project.id,
    )

    def setUp(self):
        super().setUp()

        self.cmd = network_auto_allocated_topology.CreateAutoAllocatedTopology(
            self.app, None
        )
        self.network_client.validate_auto_allocated_topology.return_value = (
            self.topology
        )

    def test_show_dry_run_no_project(self):
        arglist = [
            '--check-resources',
        ]
        verifylist = [
            ('check_resources', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        _columns, _data = self.cmd.take_action(parsed_args)

        self.network_client.validate_auto_allocated_topology.assert_called_with(
            None
        )

    def test_show_dry_run_project_option(self):
        arglist = [
            '--check-resources',
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('check_resources', True),
            ('project', self.project.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        _columns, _data = self.cmd.take_action(parsed_args)

        self.network_client.validate_auto_allocated_topology.assert_called_with(
            self.project.id
        )

    def test_show_dry_run_project_domain_option(self):
        arglist = [
            '--check-resources',
            '--project',
            self.project.id,
            '--project-domain',
            self.project.domain_id,
        ]
        verifylist = [
            ('check_resources', True),
            ('project', self.project.id),
            ('project_domain', self.project.domain_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        _columns, _data = self.cmd.take_action(parsed_args)

        self.network_client.validate_auto_allocated_topology.assert_called_with(
            self.project.id
        )


class TestDeleteAutoAllocatedTopology(TestAutoAllocatedTopology):
    project = identity_fakes.FakeProject.create_one_project()
    network_object = network_fakes.create_one_network()

    topology = network_fakes.create_one_topology(
        attrs={'id': network_object.id, 'project_id': project.id}
    )

    def setUp(self):
        super().setUp()

        self.cmd = network_auto_allocated_topology.DeleteAutoAllocatedTopology(
            self.app, None
        )
        self.network_client.delete_auto_allocated_topology.return_value = None

    def test_delete_no_project(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_auto_allocated_topology.assert_called_once_with(
            None
        )

        self.assertIsNone(result)

    def test_delete_project_arg(self):
        arglist = [
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('project', self.project.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_auto_allocated_topology.assert_called_once_with(
            self.project.id
        )

        self.assertIsNone(result)

    def test_delete_project_domain_arg(self):
        arglist = [
            '--project',
            self.project.id,
            '--project-domain',
            self.project.domain_id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('project_domain', self.project.domain_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_auto_allocated_topology.assert_called_once_with(
            self.project.id
        )

        self.assertIsNone(result)



# Copyright 2016-2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any, cast

from osc_lib.cli import identity as identity_utils
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


_attr_map = (
    ('id', 'ID', column_util.LIST_BOTH),
    ('name', 'Name', column_util.LIST_BOTH),
    ('firewall_rules', 'Firewall Rules', column_util.LIST_BOTH),
    ('description', 'Description', column_util.LIST_LONG_ONLY),
    ('audited', 'Audited', column_util.LIST_LONG_ONLY),
    ('shared', 'Shared', column_util.LIST_LONG_ONLY),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
)

_attr_map_dict = {x[0]: x[1] for x in _attr_map}


def _get_common_attrs(
    client_manager: Any,
    parsed_args: argparse.Namespace,
    is_create: bool = True,
) -> dict[str, Any]:
    attrs: dict[str, Any] = {}
    client = client_manager.network

    if parsed_args.firewall_rule and parsed_args.no_firewall_rule:
        _firewall_rules = []
        for f in parsed_args.firewall_rule:
            _firewall_rules.append(
                client.find_firewall_rule(f, ignore_missing=False).id
            )
        attrs['firewall_rules'] = _firewall_rules
    elif parsed_args.firewall_rule:
        rules = []
        if not is_create:
            fwp = client.find_firewall_policy(
                parsed_args.firewall_policy, ignore_missing=False
            )
            rules += fwp.firewall_rules
        for f in parsed_args.firewall_rule:
            rules.append(client.find_firewall_rule(f, ignore_missing=False).id)
        attrs['firewall_rules'] = rules
    elif parsed_args.no_firewall_rule:
        attrs['firewall_rules'] = []

    if parsed_args.audited is not None:
        attrs['audited'] = parsed_args.audited
    if parsed_args.name:
        attrs['name'] = parsed_args.name
    if parsed_args.description:
        attrs['description'] = parsed_args.description
    if parsed_args.shared is not None:
        attrs['shared'] = parsed_args.shared
    return attrs


def _get_common_parser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        '--description', help=_('Description of the firewall policy')
    )
    audited_group = parser.add_mutually_exclusive_group()
    audited_group.add_argument(
        '--audited',
        default=None,
        action='store_true',
        dest='audited',
        help=_('Enable auditing for the policy'),
    )
    audited_group.add_argument(
        '--no-audited',
        action='store_false',
        dest='audited',
        help=_('Disable auditing for the policy'),
    )
    shared_group = parser.add_mutually_exclusive_group()
    shared_group.add_argument(
        '--share',
        action='store_true',
        default=None,
        dest='shared',
        help=_(
            'Share the firewall policy to be used in all projects '
            '(by default, it is restricted to be used by the '
            'current project).'
        ),
    )
    shared_group.add_argument(
        '--no-share',
        action='store_false',
        dest='shared',
        help=_('Restrict use of the firewall policy to the current project'),
    )
    return parser


class CreateFirewallPolicy(command.ShowOne):
    _description = _("Create a new firewall policy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        _get_common_parser(parser)
        identity_utils.add_project_owner_option_to_parser(parser)
        parser.add_argument(
            'name', metavar='<name>', help=_('Name for the firewall policy')
        )
        fwr_group = parser.add_mutually_exclusive_group()
        fwr_group.add_argument(
            '--firewall-rule',
            action='append',
            metavar='<firewall-rule>',
            help=_('Firewall rule(s) to apply (name or ID)'),
        )
        fwr_group.add_argument(
            '--no-firewall-rule',
            action='store_true',
            help=_('Unset all firewall rules from firewall policy'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_common_attrs(self.app.client_manager, parsed_args)
        if 'project' in parsed_args and parsed_args.project is not None:
            attrs['project_id'] = identity_common.find_project(
                self.app.client_manager.identity,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
        obj = client.create_firewall_policy(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns, formatters={})
        return (display_columns, data)


class DeleteFirewallPolicy(command.Command):
    _description = _("Delete firewall policy(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall_policy',
            metavar='<firewall-policy>',
            nargs='+',
            help=_('Firewall policy(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0
        for fwp in parsed_args.firewall_policy:
            try:
                fwp_id = client.find_firewall_policy(
                    fwp, ignore_missing=False
                ).id
                client.delete_firewall_policy(fwp_id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete Firewall policy with "
                        "name or ID '%(firewall_policy)s': %(e)s"
                    ),
                    {'firewall_policy': fwp, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.firewall_policy)
            msg = _(
                "%(result)s of %(total)s firewall policy(s) failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class FirewallPolicyInsertRule(command.Command):
    _description = _("Insert a rule into a given firewall policy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall_policy',
            metavar='<firewall-policy>',
            help=_('Firewall policy to insert rule (name or ID)'),
        )
        parser.add_argument(
            '--insert-before',
            metavar='<firewall-rule>',
            help=_(
                'Insert the new rule before this existing rule  (name or ID)'
            ),
        )
        parser.add_argument(
            '--insert-after',
            metavar='<firewall-rule>',
            help=_(
                'Insert the new rule after this existing rule  (name or ID)'
            ),
        )
        parser.add_argument(
            'firewall_rule',
            metavar='<firewall-rule>',
            help=_('Firewall rule to be inserted (name or ID)'),
        )
        return parser

    def args2body(self, parsed_args: argparse.Namespace) -> dict[str, str]:
        client = self.app.client_manager.network
        _rule_id = _get_required_firewall_rule(client, parsed_args)
        _insert_before = ''
        if 'insert_before' in parsed_args:
            if parsed_args.insert_before:
                _insert_before = client.find_firewall_rule(
                    parsed_args.insert_before, ignore_missing=False
                ).id
        _insert_after = ''
        if 'insert_after' in parsed_args:
            if parsed_args.insert_after:
                _insert_after = client.find_firewall_rule(
                    parsed_args.insert_after, ignore_missing=False
                ).id
        return {
            'firewall_rule_id': _rule_id,
            'insert_before': _insert_before,
            'insert_after': _insert_after,
        }

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        policy_id = client.find_firewall_policy(
            parsed_args.firewall_policy, ignore_missing=False
        ).id
        body = self.args2body(parsed_args)
        client.insert_rule_into_policy(policy_id, **body)
        rule_id = body['firewall_rule_id']
        policy = parsed_args.firewall_policy
        print(
            (
                _(
                    'Inserted firewall rule %(rule)s in firewall policy '
                    '%(policy)s'
                )
                % {'rule': rule_id, 'policy': policy}
            ),
            file=self.app.stdout,
        )


class FirewallPolicyRemoveRule(command.Command):
    _description = _("Remove a rule from a given firewall policy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall_policy',
            metavar='<firewall-policy>',
            help=_('Firewall policy to remove rule (name or ID)'),
        )
        parser.add_argument(
            'firewall_rule',
            metavar='<firewall-rule>',
            help=_('Firewall rule to remove from policy (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        policy_id = client.find_firewall_policy(
            parsed_args.firewall_policy, ignore_missing=False
        ).id
        fwr_id = _get_required_firewall_rule(client, parsed_args)
        body = {'firewall_rule_id': fwr_id}
        client.remove_rule_from_policy(policy_id, **body)
        rule_id = body['firewall_rule_id']
        policy = parsed_args.firewall_policy
        print(
            (
                _(
                    'Removed firewall rule %(rule)s from firewall policy '
                    '%(policy)s'
                )
                % {'rule': rule_id, 'policy': policy}
            ),
            file=self.app.stdout,
        )


class ListFirewallPolicy(command.Lister):
    _description = _("List firewall policies")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        obj = client.firewall_policies()
        headers, columns = column_util.get_column_definitions(
            list(_attr_map), long_listing=parsed_args.long
        )
        return (
            headers,
            (
                utils.get_dict_properties(s, columns, formatters={})
                for s in obj
            ),
        )


class SetFirewallPolicy(command.Command):
    _description = _("Set firewall policy properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        _get_common_parser(parser)
        parser.add_argument(
            'firewall_policy',
            metavar='<firewall-policy>',
            help=_('Firewall policy to update (name or ID)'),
        )
        parser.add_argument(
            '--name', metavar='<name>', help=_('Name for the firewall policy')
        )
        parser.add_argument(
            '--firewall-rule',
            action='append',
            metavar='<firewall-rule>',
            help=_('Firewall rule(s) to apply (name or ID)'),
        )
        parser.add_argument(
            '--no-firewall-rule',
            action='store_true',
            help=_('Remove all firewall rules from firewall policy'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        fwp_id = client.find_firewall_policy(
            parsed_args.firewall_policy, ignore_missing=False
        ).id
        attrs = _get_common_attrs(
            self.app.client_manager, parsed_args, is_create=False
        )
        try:
            client.update_firewall_policy(fwp_id, **attrs)
        except Exception as e:
            msg = _("Failed to set firewall policy '%(policy)s': %(e)s") % {
                'policy': parsed_args.firewall_policy,
                'e': e,
            }
            raise exceptions.CommandError(msg)


class ShowFirewallPolicy(command.ShowOne):
    _description = _("Display firewall policy details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall_policy',
            metavar='<firewall-policy>',
            help=_('Firewall policy to show (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        fwp_id = client.find_firewall_policy(
            parsed_args.firewall_policy, ignore_missing=False
        ).id
        obj = client.get_firewall_policy(fwp_id)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns, formatters={})
        return (display_columns, data)


def _get_required_firewall_rule(
    client: Any, parsed_args: argparse.Namespace
) -> str:
    if not parsed_args.firewall_rule:
        msg = _("Firewall rule (name or ID) is required.")
        raise exceptions.CommandError(msg)
    return cast(
        str,
        client.find_firewall_rule(
            parsed_args.firewall_rule, ignore_missing=False
        ).id,
    )


class UnsetFirewallPolicy(command.Command):
    _description = _("Unset firewall policy properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall_policy',
            metavar='<firewall-policy>',
            help=_('Firewall policy to unset (name or ID)'),
        )
        firewall_rule_group = parser.add_mutually_exclusive_group()
        firewall_rule_group.add_argument(
            '--firewall-rule',
            action='append',
            metavar='<firewall-rule>',
            help=_(
                'Remove firewall rule(s) from the firewall policy (name or ID)'
            ),
        )
        firewall_rule_group.add_argument(
            '--all-firewall-rule',
            action='store_true',
            help=_('Remove all firewall rules from the firewall policy'),
        )
        parser.add_argument(
            '--audited',
            action='store_true',
            help=_('Disable auditing for the policy'),
        )
        parser.add_argument(
            '--share',
            action='store_true',
            help=_(
                '(Deprecated) Use "firewall policy set --no-share" instead. '
                'Restrict use of the firewall policy to the current project'
            ),
        )
        return parser

    def _get_attrs(
        self, client_manager: Any, parsed_args: argparse.Namespace
    ) -> dict[str, Any]:
        attrs: dict[str, Any] = {}
        client = client_manager.network

        if parsed_args.firewall_rule:
            current = client.find_firewall_policy(
                parsed_args.firewall_policy, ignore_missing=False
            ).firewall_rules
            removed = []
            for f in set(parsed_args.firewall_rule):
                removed.append(
                    client.find_firewall_rule(f, ignore_missing=False).id
                )
            attrs['firewall_rules'] = [r for r in current if r not in removed]
        if parsed_args.all_firewall_rule:
            attrs['firewall_rules'] = []
        if parsed_args.audited:
            attrs['audited'] = False
        if parsed_args.share:
            LOG.warning(
                'The --share option is deprecated, please use '
                '"firewall policy set --no-share" instead.'
            )
            attrs['shared'] = False
        return attrs

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        fwp_id = client.find_firewall_policy(
            parsed_args.firewall_policy, ignore_missing=False
        ).id
        attrs = self._get_attrs(self.app.client_manager, parsed_args)
        try:
            client.update_firewall_policy(fwp_id, **attrs)
        except Exception as e:
            msg = _("Failed to unset firewall policy '%(policy)s': %(e)s") % {
                'policy': parsed_args.firewall_policy,
                'e': e,
            }
            raise exceptions.CommandError(msg)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import random
import re
from unittest import mock
import uuid

from keystoneauth1 import discover
from openstack.compute.v2 import _proxy
from openstack.compute.v2 import availability_zone as _availability_zone
from openstack.compute.v2 import extension as _extension
from openstack.compute.v2 import flavor as _flavor
from openstack.compute.v2 import limits as _limits
from openstack.compute.v2 import migration as _migration
from openstack.compute.v2 import server as _server
from openstack.compute.v2 import server_action as _server_action
from openstack.compute.v2 import server_interface as _server_interface
from openstack.compute.v2 import server_migration as _server_migration
from openstack.compute.v2 import server_share as _server_share
from openstack.compute.v2 import volume_attachment as _volume_attachment
from openstack.test import fakes as sdk_fakes

from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit.image.v2 import fakes as image_fakes
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        self.app.client_manager.compute = mock.Mock(_proxy.Proxy)
        self.compute_client = self.app.client_manager.compute
        self.set_compute_api_version()  # default to the lowest

    def set_compute_api_version(self, version: str = '2.1'):
        """Set a fake server version.

        :param version: The fake microversion to "support". This should be a
            string of format '2.xx'.
        :returns: None
        """
        assert re.match(r'2.\d+', version)

        self.compute_client.default_microversion = version
        self.compute_client.get_endpoint_data.return_value = (
            discover.EndpointData(
                min_microversion='2.1',  # nova has not bumped this yet
                max_microversion=version,
            )
        )


class TestComputev2(
    identity_fakes.FakeClientMixin,
    network_fakes.FakeClientMixin,
    image_fakes.FakeClientMixin,
    volume_fakes.FakeClientMixin,
    FakeClientMixin,
    utils.TestCommand,
): ...


def create_one_agent(attrs=None):
    """Create a fake agent.

    :param dict attrs: A dictionary with all attributes
    :return: A dicionarty faking the agent
    """

    attrs = attrs or {}

    # set default attributes.
    agent_attrs = {
        'agent_id': 'agent-id-' + uuid.uuid4().hex,
        'os': 'agent-os-' + uuid.uuid4().hex,
        'architecture': 'agent-architecture',
        'version': '8.0',
        'url': 'http://127.0.0.1',
        'md5hash': 'agent-md5hash',
        'hypervisor': 'hypervisor',
    }

    assert not set(attrs) - set(agent_attrs), 'unknown keys'

    # Overwrite default attributes.
    agent_attrs.update(attrs)

    return agent_attrs


def create_agents(attrs=None, count=2):
    """Create multiple fake agents.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of agents to fake
    :return: A list of dictionaries faking the agents
    """
    agents = []
    for i in range(0, count):
        agents.append(create_one_agent(attrs))

    return agents


def create_one_extension(attrs=None):
    """Create a fake extension.

    :param dict attrs: A dictionary with all attributes
    :return: A fake :class:`~openstack.compute.v2.extension.Extension` object
    """
    attrs = attrs or {}

    # Set default attributes.
    extension_info = {
        'alias': 'NMN',
        'description': 'description-' + uuid.uuid4().hex,
        'links': [
            {
                "href": "https://github.com/openstack/compute-api",
                "type": "text/html",
                "rel": "describedby",
            }
        ],
        'name': 'name-' + uuid.uuid4().hex,
        'namespace': (
            'http://docs.openstack.org/compute/ext/multinic/api/v1.1'
        ),
        'updated_at': '2014-01-07T12:00:0-00:00',
    }

    # Overwrite default attributes.
    extension_info.update(attrs)

    extension = _extension.Extension(**extension_info)
    return extension


def create_one_security_group(attrs=None):
    """Create a fake security group.

    :param dict attrs: A dictionary with all attributes
    :return: A dictionary faking the security group
    """
    attrs = attrs or {}

    # Set default attributes.
    security_group_attrs = {
        'id': 'security-group-id-' + uuid.uuid4().hex,
        'name': 'security-group-name-' + uuid.uuid4().hex,
        'description': 'security-group-description-' + uuid.uuid4().hex,
        'tenant_id': 'project-id-' + uuid.uuid4().hex,
        'rules': [],
    }

    assert not set(attrs) - set(security_group_attrs), 'unknown keys'

    # Overwrite default attributes.
    security_group_attrs.update(attrs)
    return security_group_attrs


def create_security_groups(attrs=None, count=2):
    """Create multiple fake security groups.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of security groups to fake
    :return: A list of dictionaries faking the security groups
    """
    security_groups = []
    for i in range(0, count):
        security_groups.append(create_one_security_group(attrs))

    return security_groups


def create_one_security_group_rule(attrs=None):
    """Create a fake security group rule.

    :param dict attrs: A dictionary with all attributes
    :return: A dictionary faking the security group rule
    """
    attrs = attrs or {}

    # Set default attributes.
    security_group_rule_attrs = {
        'from_port': 0,
        'group': {},
        'id': 'security-group-rule-id-' + uuid.uuid4().hex,
        'ip_protocol': 'tcp',
        'ip_range': {'cidr': '0.0.0.0/0'},
        'parent_group_id': 'security-group-id-' + uuid.uuid4().hex,
        'to_port': 0,
    }

    assert not set(attrs) - set(security_group_rule_attrs), 'unknown keys'

    # Overwrite default attributes.
    security_group_rule_attrs.update(attrs)

    return security_group_rule_attrs


def create_security_group_rules(attrs=None, count=2):
    """Create multiple fake security group rules.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of security group rules to fake
    :return: A list of dictionaries faking the security group rules
    """
    security_group_rules = []
    for i in range(0, count):
        security_group_rules.append(create_one_security_group_rule(attrs))

    return security_group_rules


def create_one_server(attrs=None):
    """Create a fake server

    :param dict attrs: A dictionary with all attributes
    :return: A fake :class:`~openstack.compute.v2.server.Server` object,
    """
    attrs = attrs or {}

    # Set default attributes.
    server_info = {
        'id': 'server-id-' + uuid.uuid4().hex,
        'name': 'server-name-' + uuid.uuid4().hex,
        'metadata': {},
        'image': {
            'id': 'image-id-' + uuid.uuid4().hex,
        },
        'flavor': {
            'id': 'flavor-id-' + uuid.uuid4().hex,
        },
        'OS-EXT-STS:power_state': 1,
    }

    # Overwrite default attributes.
    server_info.update(attrs)
    server = _server.Server(**server_info)

    # Override methods
    server.trigger_crash_dump = mock.MagicMock()

    return server


def create_servers(attrs=None, count=2):
    """Create multiple fake servers

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of servers to fake
    :return: A list of fake :class:`openstack.compute.v2.server.Server` objects
    """
    servers = []
    for i in range(0, count):
        servers.append(create_one_server(attrs))

    return servers


def create_one_server_action(attrs=None):
    """Create a fake server action.

    :param attrs: A dictionary with all attributes
    :return: A fake :class:`~openstack.compute.v2.server_action.ServerAction`
        object
    """
    attrs = attrs or {}

    # Set default attributes
    server_action_info = {
        "server_id": "server-event-" + uuid.uuid4().hex,
        "user_id": "user-id-" + uuid.uuid4().hex,
        "start_time": "2017-02-27T07:47:13.000000",
        "request_id": "req-" + uuid.uuid4().hex,
        "action": "create",
        "message": None,
        "project_id": "project-id-" + uuid.uuid4().hex,
        "events": [
            {
                "finish_time": "2017-02-27T07:47:25.000000",
                "start_time": "2017-02-27T07:47:15.000000",
                "traceback": None,
                "event": "compute__do_build_and_run_instance",
                "result": "Success",
            }
        ],
    }
    # Overwrite default attributes
    server_action_info.update(attrs)

    # We handle events separately since they're nested resources
    events = [
        _server_action.ServerActionEvent(**event)
        for event in server_action_info.pop('events')
    ]

    server_action = _server_action.ServerAction(
        **server_action_info,
        events=events,
    )
    return server_action


def create_one_flavor(attrs=None):
    """Create a fake flavor.

    :param dict attrs: A dictionary with all attributes
    :return: A fake :class:`~openstack.compute.v2.flavor.Flavor` object
    """
    attrs = attrs or {}

    # Set default attributes.
    flavor_info = {
        'id': 'flavor-id-' + uuid.uuid4().hex,
        'name': 'flavor-name-' + uuid.uuid4().hex,
        'ram': 8192,
        'vcpus': 4,
        'disk': 128,
        'swap': 0,
        'rxtx_factor': 1.0,
        'OS-FLV-DISABLED:disabled': False,
        'os-flavor-access:is_public': True,
        'description': 'description',
        'OS-FLV-EXT-DATA:ephemeral': 0,
        'extra_specs': {'property': 'value'},
    }

    # Overwrite default attributes.
    flavor_info.update(attrs)

    flavor = _flavor.Flavor(**flavor_info)

    return flavor


def create_flavors(attrs=None, count=2):
    """Create multiple fake flavors.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of flavors to fake
    :return: A list of fake :class:`openstack.compute.v2.flavor.Flavor` objects
    """
    flavors = []
    for i in range(0, count):
        flavors.append(create_one_flavor(attrs))

    return flavors


def create_one_flavor_access(attrs=None):
    """Create a fake flavor access.

    :param dict attrs: A dictionary with all attributes
    :return: A dictionary faking the flavor access
    """
    attrs = attrs or {}

    # Set default attributes.
    flavor_access_info = {
        'flavor_id': 'flavor-id-' + uuid.uuid4().hex,
        'tenant_id': 'tenant-id-' + uuid.uuid4().hex,
    }

    assert not set(attrs) - set(flavor_access_info), 'unknown keys'

    # Overwrite default attributes.
    flavor_access_info.update(attrs)

    return flavor_access_info


def create_one_availability_zone(attrs=None):
    """Create a fake AZ.

    :param dict attrs: A dictionary with all attributes
    :return: A fake
        :class:`~openstack.compute.v2.availability_zone.AvailabilityZone` object
    """
    attrs = attrs or {}

    # Set default attributes.
    host_name = uuid.uuid4().hex
    service_name = uuid.uuid4().hex
    availability_zone_info = {
        'name': uuid.uuid4().hex,
        'state': {'available': True},
        'hosts': {
            host_name: {
                service_name: {
                    'available': True,
                    'active': True,
                    'updated_at': '2023-01-01T00:00:00.000000',
                }
            }
        },
    }

    # Overwrite default attributes.
    availability_zone_info.update(attrs)

    availability_zone = _availability_zone.AvailabilityZone(
        **availability_zone_info
    )
    return availability_zone


def create_availability_zones(attrs=None, count=2):
    """Create multiple fake AZs.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of availability zones to fake
    :return: A list of fake
        openstack.compute.v2.availability_zone.AvailabilityZone objects
    """
    availability_zones = []
    for i in range(0, count):
        availability_zone = create_one_availability_zone(attrs)
        availability_zones.append(availability_zone)

    return availability_zones


def create_one_floating_ip(attrs=None):
    """Create a fake floating IP.

    :param dict attrs: A dictionary with all attributes
    :return: A dictionary faking the floating IP
    """
    attrs = attrs or {}

    # Set default attributes.
    floating_ip_attrs = {
        'id': 'floating-ip-id-' + uuid.uuid4().hex,
        'ip': '1.0.9.0',
        'fixed_ip': '2.0.9.0',
        'instance_id': 'server-id-' + uuid.uuid4().hex,
        'pool': 'public',
    }

    assert not set(attrs) - set(floating_ip_attrs), 'unknown keys'

    # Overwrite default attributes.
    floating_ip_attrs.update(attrs)

    return floating_ip_attrs


def create_floating_ips(attrs=None, count=2):
    """Create multiple fake floating IPs.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of floating IPs to fake
    :return: A list of dictionaries faking the floating IPs
    """
    floating_ips = []
    for i in range(0, count):
        floating_ips.append(create_one_floating_ip(attrs))
    return floating_ips


def create_one_floating_ip_pool(attrs=None):
    """Create a fake floating IP pool.

    :param dict attrs: A dictionary with all attributes
    :return: A dictionary faking the floating IP pool
    """
    if attrs is None:
        attrs = {}

    # Set default attributes.
    floating_ip_pool_attrs = {
        'name': 'floating-ip-pool-name-' + uuid.uuid4().hex,
    }

    assert not set(attrs) - set(floating_ip_pool_attrs), 'unknown keys'

    # Overwrite default attributes.
    floating_ip_pool_attrs.update(attrs)

    return floating_ip_pool_attrs


def create_floating_ip_pools(attrs=None, count=2):
    """Create multiple fake floating IP pools.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of floating IP pools to fake
    :return: A list of dictionaries faking the floating IP pools
    """
    floating_ip_pools = []
    for i in range(0, count):
        floating_ip_pools.append(create_one_floating_ip_pool(attrs))
    return floating_ip_pools


def create_one_network(attrs=None):
    """Create a fake network.

    :param dict attrs: A dictionary with all attributes
    :return: A dictionary faking the network
    """
    attrs = attrs or {}

    # Set default attributes.
    network_attrs = {
        'bridge': 'br100',
        'bridge_interface': None,
        'broadcast': '10.0.0.255',
        'cidr': '10.0.0.0/24',
        'cidr_v6': None,
        'created_at': '2016-02-11T11:17:37.000000',
        'deleted': False,
        'deleted_at': None,
        'dhcp_server': '10.0.0.1',
        'dhcp_start': '10.0.0.2',
        'dns1': '8.8.4.4',
        'dns2': None,
        'enable_dhcp': True,
        'gateway': '10.0.0.1',
        'gateway_v6': None,
        'host': None,
        'id': 'network-id-' + uuid.uuid4().hex,
        'injected': False,
        'label': 'network-label-' + uuid.uuid4().hex,
        'mtu': None,
        'multi_host': False,
        'netmask': '255.255.255.0',
        'netmask_v6': None,
        'priority': None,
        'project_id': 'project-id-' + uuid.uuid4().hex,
        'rxtx_base': None,
        'share_address': False,
        'updated_at': None,
        'vlan': None,
        'vpn_private_address': None,
        'vpn_public_address': None,
        'vpn_public_port': None,
    }

    assert not set(attrs) - set(network_attrs), 'unknown keys'

    # Overwrite default attributes.
    network_attrs.update(attrs)

    return network_attrs


def create_networks(attrs=None, count=2):
    """Create multiple fake networks.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of networks to fake
    :return: A list of dictionaries faking the networks
    """
    networks = []
    for i in range(0, count):
        networks.append(create_one_network(attrs))

    return networks


def create_limits(attrs=None):
    """Create a fake limits object."""
    attrs = attrs or {}

    limits_attrs = {
        'absolute': {
            'maxServerMeta': 128,
            'maxTotalInstances': 10,
            'maxPersonality': 5,
            'totalServerGroupsUsed': 0,
            'maxImageMeta': 128,
            'maxPersonalitySize': 10240,
            'maxTotalRAMSize': 51200,
            'maxServerGroups': 10,
            'maxSecurityGroupRules': 20,
            'maxTotalKeypairs': 100,
            'totalCoresUsed': 0,
            'totalRAMUsed': 0,
            'maxSecurityGroups': 10,
            'totalFloatingIpsUsed': 0,
            'totalInstancesUsed': 0,
            'maxServerGroupMembers': 10,
            'maxTotalFloatingIps': 10,
            'totalSecurityGroupsUsed': 0,
            'maxTotalCores': 20,
        },
        'rate': [
            {
                "uri": "*",
                "limit": [
                    {
                        "value": 10,
                        "verb": "POST",
                        "remaining": 2,
                        "unit": "MINUTE",
                        "next-available": "2011-12-15T22:42:45Z",
                    },
                    {
                        "value": 10,
                        "verb": "PUT",
                        "remaining": 2,
                        "unit": "MINUTE",
                        "next-available": "2011-12-15T22:42:45Z",
                    },
                    {
                        "value": 100,
                        "verb": "DELETE",
                        "remaining": 100,
                        "unit": "MINUTE",
                        "next-available": "2011-12-15T22:42:45Z",
                    },
                ],
            }
        ],
    }
    limits_attrs.update(attrs)

    limits = _limits.Limits(**limits_attrs)
    return limits


def create_one_migration(attrs=None):
    """Create a fake migration.

    :param dict attrs: A dictionary with all attributes
    :return: A fake :class:`~openstack.compute.v2.migration.Migration` object
    """
    attrs = attrs or {}

    # Set default attributes.
    migration_info = {
        "created_at": "2017-01-31T08:03:21.000000",
        "dest_compute": "compute-" + uuid.uuid4().hex,
        "dest_host": "10.0.2.15",
        "dest_node": "node-" + uuid.uuid4().hex,
        "id": random.randint(1, 999),
        "migration_type": "migration",
        "new_flavor_id": uuid.uuid4().hex,
        "old_flavor_id": uuid.uuid4().hex,
        "project_id": uuid.uuid4().hex,
        "server_id": uuid.uuid4().hex,
        "source_compute": "compute-" + uuid.uuid4().hex,
        "source_node": "node-" + uuid.uuid4().hex,
        "status": "migrating",
        "updated_at": "2017-01-31T08:03:25.000000",
        "user_id": uuid.uuid4().hex,
        "uuid": uuid.uuid4().hex,
    }

    # Overwrite default attributes.
    migration_info.update(attrs)

    migration = _migration.Migration(**migration_info)
    return migration


def create_migrations(attrs=None, count=2):
    """Create multiple fake migrations.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of migrations to fake
    :return: A list of fake :class:`openstack.compute.v2.migration.Migration` objects
    """
    migrations = []
    for i in range(0, count):
        migrations.append(create_one_migration(attrs))

    return migrations


def create_one_server_migration(attrs=None):
    """Create a fake server migration.

    :param dict attrs: A dictionary with all attributes
    :return: A fake
        :class:`~openstack.compute.v2.server_migration.ServerMigration` object
    """
    attrs = attrs or {}

    # Set default attributes.

    migration_info = {
        "created_at": "2016-01-29T13:42:02.000000",
        "dest_compute": "compute2",
        "dest_host": "1.2.3.4",
        "dest_node": "node2",
        "id": random.randint(1, 999),
        "server_uuid": uuid.uuid4().hex,
        "source_compute": "compute1",
        "source_node": "node1",
        "status": "running",
        "memory_total_bytes": random.randint(1, 99999),
        "memory_processed_bytes": random.randint(1, 99999),
        "memory_remaining_bytes": random.randint(1, 99999),
        "disk_total_bytes": random.randint(1, 99999),
        "disk_processed_bytes": random.randint(1, 99999),
        "disk_remaining_bytes": random.randint(1, 99999),
        "updated_at": "2016-01-29T13:42:02.000000",
        # added in 2.59
        "uuid": uuid.uuid4().hex,
        # added in 2.80
        "user_id": uuid.uuid4().hex,
        "project_id": uuid.uuid4().hex,
    }

    # Overwrite default attributes.
    migration_info.update(attrs)

    migration = _server_migration.ServerMigration(**migration_info)
    return migration


def create_server_migrations(attrs=None, methods=None, count=2):
    """Create multiple server migrations.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of server migrations to fake
    :return A list of fake
        openstack.compute.v2.server_migration.ServerMigration objects
    """
    migrations = []
    for i in range(0, count):
        migrations.append(create_one_server_migration(attrs, methods))

    return migrations


def create_one_volume_attachment(attrs=None):
    """Create a fake volume attachment.

    :param dict attrs: A dictionary with all attributes
    :return: A fake
        :class:`~openstack.compute.v2.volume_attachment.VolumeAttachment` object
    """
    attrs = attrs or {}

    # Set default attributes.
    volume_attachment_info = {
        "id": uuid.uuid4().hex,
        "device": "/dev/sdb",
        "server_id": uuid.uuid4().hex,
        "volume_id": uuid.uuid4().hex,
        # introduced in API microversion 2.70
        "tag": "foo",
        # introduced in API microversion 2.79
        "delete_on_termination": True,
        # introduced in API microversion 2.89
        "attachment_id": uuid.uuid4().hex,
        "bdm_id": uuid.uuid4().hex,
    }

    # Overwrite default attributes.
    volume_attachment_info.update(attrs)

    return _volume_attachment.VolumeAttachment(**volume_attachment_info)


def create_volume_attachments(attrs=None, count=2):
    """Create multiple fake volume attachments.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of volume attachments to fake
    :return: A list of fake
        openstack.compute.v2.volume_attachment.VolumeAttachment objects
    """
    volume_attachments = []
    for i in range(0, count):
        volume_attachments.append(create_one_volume_attachment(attrs))

    return volume_attachments


def create_one_share(attrs=None):
    """Create a fake share attachment.

    :param dict attrs: A dictionary with all attributes
    :return: A fake openstack.compute.v2.server_share.ShareMapping
        object
    """
    return sdk_fakes.generate_fake_resource(
        _server_share.ShareMapping, **(attrs or {})
    )


def create_shares(attrs=None, count=2):
    """Create multiple fake share attachments.

    :param dict attrs: A dictionary with all attributes
    :param int count: The number of share attachments to fake
    :return: A list of fake
        openstack.compute.v2.server_share.ShareMapping objects
    """
    return [create_one_share(attrs) for _ in range(count)]


def create_one_server_interface(attrs=None):
    """Create a fake ServerInterface.

    :param dict attrs: A dictionary with all attributes
    :param dict methods: A dictionary with all methods
    :return: A fake
        :class:`~openstack.compute.v2.server_interface.ServerInterface` object
    """
    attrs = attrs or {}

    # Set default attributes.
    server_interface_info = {
        "fixed_ips": uuid.uuid4().hex,
        "mac_addr": "aa:aa:aa:aa:aa:aa",
        "net_id": uuid.uuid4().hex,
        "port_id": uuid.uuid4().hex,
        "port_state": "ACTIVE",
        "server_id": uuid.uuid4().hex,
        # introduced in API microversion 2.70
        "tag": "foo",
    }

    # Overwrite default attributes.
    server_interface_info.update(attrs)

    return _server_interface.ServerInterface(**server_interface_info)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.compute.v2 import console_auth_token as _console_auth_token
from openstack.test import fakes as sdk_fakes

from openstackclient.compute.v2 import console_connection
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestConsoleTokens(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._console_auth_token = sdk_fakes.generate_fake_resource(
            _console_auth_token.ConsoleAuthToken,
            host='127.0.0.1',
            instance_uuid=uuid.uuid4().hex,
            internal_access_path=None,
            port=5900,
            tls_port=5901,
        )
        self.compute_client.validate_console_auth_token.return_value = (
            self._console_auth_token
        )

        self.columns = (
            'host',
            'instance_uuid',
            'internal_access_path',
            'port',
            'tls_port',
        )
        self.data = (
            self._console_auth_token.host,
            self._console_auth_token.instance_uuid,
            self._console_auth_token.internal_access_path,
            self._console_auth_token.port,
            self._console_auth_token.tls_port,
        )

        self.cmd = console_connection.ShowConsoleConnectionInformation(
            self.app, None
        )

    def test_console_connection_show(self):
        arglist = [
            'token',
        ]
        verifylist = [
            ('token', 'token'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.validate_console_auth_token.assert_called_once_with(
            'token'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from osc_lib import exceptions

from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import role as _role
from openstack.identity.v3 import trust as _trust
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes

from openstackclient.identity.v3 import trust
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestTrustCreate(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.trust = sdk_fakes.generate_fake_resource(_trust.Trust)
        self.identity_sdk_client.create_trust.return_value = self.trust

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.user = sdk_fakes.generate_fake_resource(_user.User)
        self.identity_sdk_client.find_user.return_value = self.user

        self.role = sdk_fakes.generate_fake_resource(_role.Role)
        self.identity_sdk_client.find_role.return_value = self.role

        # Get the command object to test
        self.cmd = trust.CreateTrust(self.app, None)

    def test_trust_create_basic(self):
        arglist = [
            '--project',
            self.project.id,
            '--role',
            self.role.id,
            self.user.id,
            self.user.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('roles', [self.role.id]),
            ('trustor', self.user.id),
            ('trustee', self.user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'project_id': self.project.id,
            'roles': [{'id': self.role.id}],
            'impersonation': False,
        }
        # TrustManager.create(trustee_id, trustor_id, impersonation=,
        #   project=, role_names=, expires_at=)
        self.identity_sdk_client.create_trust.assert_called_with(
            trustor_user_id=self.user.id,
            trustee_user_id=self.user.id,
            **kwargs,
        )

        collist = (
            'expires_at',
            'id',
            'is_impersonation',
            'project_id',
            'redelegated_trust_id',
            'redelegation_count',
            'remaining_uses',
            'roles',
            'trustee_user_id',
            'trustor_user_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.trust.expires_at,
            self.trust.id,
            self.trust.is_impersonation,
            self.trust.project_id,
            self.trust.redelegated_trust_id,
            self.trust.redelegation_count,
            self.trust.remaining_uses,
            self.trust.roles,
            self.trust.trustee_user_id,
            self.trust.trustor_user_id,
        )
        self.assertEqual(datalist, data)


class TestTrustDelete(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.trust = sdk_fakes.generate_fake_resource(_trust.Trust)
        self.identity_sdk_client.delete_trust.return_value = None
        self.identity_sdk_client.find_trust.return_value = self.trust

        # Get the command object to test
        self.cmd = trust.DeleteTrust(self.app, None)

    def test_trust_delete(self):
        arglist = [
            self.trust.id,
        ]
        verifylist = [('trust', [self.trust.id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_trust.assert_called_with(
            self.trust.id,
        )
        self.assertIsNone(result)

    def test_delete_multi_trusts_with_exception(self):
        arglist = [
            self.trust.id,
            'unexist_trust',
        ]
        verifylist = [
            ('trust', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.identity_sdk_client.find_trust.side_effect = [
            self.trust,
            sdk_exceptions.ResourceNotFound,
        ]

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 trusts failed to delete.', str(e))

        self.identity_sdk_client.find_trust.assert_has_calls(
            [
                mock.call(self.trust.id, ignore_missing=False),
                mock.call('unexist_trust', ignore_missing=False),
            ]
        )
        self.identity_sdk_client.delete_trust.assert_called_once_with(
            self.trust.id
        )


class TestTrustList(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.trust = sdk_fakes.generate_fake_resource(_trust.Trust)
        self.identity_sdk_client.trusts.return_value = [self.trust]

        self.user = sdk_fakes.generate_fake_resource(_user.User)
        self.identity_sdk_client.find_user.return_value = self.user

        # Get the command object to test
        self.cmd = trust.ListTrust(self.app, None)

    def test_trust_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.trusts.assert_called_with(
            trustor_user_id=None,
            trustee_user_id=None,
        )

        collist = (
            'ID',
            'Expires At',
            'Impersonation',
            'Project ID',
            'Trustee User ID',
            'Trustor User ID',
        )
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.trust.id,
                self.trust.expires_at,
                self.trust.is_impersonation,
                self.trust.project_id,
                self.trust.trustee_user_id,
                self.trust.trustor_user_id,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_trust_list_auth_user(self):
        self.app.client_manager.auth_ref = mock.Mock()
        auth_ref = self.app.client_manager.auth_ref

        arglist = ['--auth-user']
        verifylist = [
            ('trustor', None),
            ('trustee', None),
            ('authuser', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.trusts.assert_has_calls(
            [
                mock.call(trustor_user_id=auth_ref.user_id),
                mock.call(trustee_user_id=auth_ref.user_id),
            ]
        )

        collist = (
            'ID',
            'Expires At',
            'Impersonation',
            'Project ID',
            'Trustee User ID',
            'Trustor User ID',
        )
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.trust.id,
                self.trust.expires_at,
                self.trust.is_impersonation,
                self.trust.project_id,
                self.trust.trustee_user_id,
                self.trust.trustor_user_id,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_trust_list_trustee(self):
        arglist = ['--trustee', self.user.name]
        verifylist = [
            ('trustor', None),
            ('trustee', self.user.name),
            ('authuser', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.trusts.assert_called_with(
            trustee_user_id=self.user.id,
            trustor_user_id=None,
        )

        collist = (
            'ID',
            'Expires At',
            'Impersonation',
            'Project ID',
            'Trustee User ID',
            'Trustor User ID',
        )
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.trust.id,
                self.trust.expires_at,
                self.trust.is_impersonation,
                self.trust.project_id,
                self.trust.trustee_user_id,
                self.trust.trustor_user_id,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_trust_list_trustor(self):
        arglist = ['--trustor', self.user.name]
        verifylist = [
            ('trustee', None),
            ('trustor', self.user.name),
            ('authuser', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.trusts.assert_called_once_with(
            trustor_user_id=self.user.id,
            trustee_user_id=None,
        )

        collist = (
            'ID',
            'Expires At',
            'Impersonation',
            'Project ID',
            'Trustee User ID',
            'Trustor User ID',
        )
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.trust.id,
                self.trust.expires_at,
                self.trust.is_impersonation,
                self.trust.project_id,
                self.trust.trustee_user_id,
                self.trust.trustor_user_id,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestTrustShow(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.trust = sdk_fakes.generate_fake_resource(_trust.Trust)
        self.identity_sdk_client.find_trust.return_value = self.trust

        # Get the command object to test
        self.cmd = trust.ShowTrust(self.app, None)

    def test_trust_show(self):
        arglist = [
            self.trust.id,
        ]
        verifylist = [
            ('trust', self.trust.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_trust.assert_called_with(
            self.trust.id, ignore_missing=False
        )

        collist = (
            'expires_at',
            'id',
            'is_impersonation',
            'project_id',
            'redelegated_trust_id',
            'redelegation_count',
            'remaining_uses',
            'roles',
            'trustee_user_id',
            'trustor_user_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.trust.expires_at,
            self.trust.id,
            self.trust.is_impersonation,
            self.trust.project_id,
            self.trust.redelegated_trust_id,
            self.trust.redelegation_count,
            self.trust.remaining_uses,
            self.trust.roles,
            self.trust.trustee_user_id,
            self.trust.trustor_user_id,
        )
        self.assertEqual(datalist, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from openstack.network.v2 import floating_ip as _floating_ip
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.network.v2 import floating_ip as fip
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestFloatingIPNetwork(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains


class TestCreateFloatingIPNetwork(TestFloatingIPNetwork):
    # Fake data for option tests.
    floating_network = network_fakes.create_one_network()
    subnet = network_fakes.FakeSubnet.create_one_subnet()
    port = network_fakes.create_one_port()

    # The floating ip created.
    floating_ip = network_fakes.FakeFloatingIP.create_one_floating_ip(
        attrs={
            'floating_network_id': floating_network.id,
            'port_id': port.id,
            'dns_domain': 'example.org.',
            'dns_name': 'fip1',
        }
    )

    columns = (
        'description',
        'dns_domain',
        'dns_name',
        'fixed_ip_address',
        'floating_ip_address',
        'floating_network_id',
        'id',
        'port_id',
        'project_id',
        'qos_policy_id',
        'router_id',
        'status',
        'tags',
    )

    data = (
        floating_ip.description,
        floating_ip.dns_domain,
        floating_ip.dns_name,
        floating_ip.fixed_ip_address,
        floating_ip.floating_ip_address,
        floating_ip.floating_network_id,
        floating_ip.id,
        floating_ip.port_id,
        floating_ip.project_id,
        floating_ip.qos_policy_id,
        floating_ip.router_id,
        floating_ip.status,
        floating_ip.tags,
    )

    def setUp(self):
        super().setUp()

        self.network_client.create_ip.return_value = self.floating_ip

        self.network_client.set_tags.return_value = None

        self.network_client.find_network.return_value = self.floating_network

        self.network_client.find_subnet.return_value = self.subnet
        self.network_client.find_port.return_value = self.port

        # Get the command object to test
        self.cmd = fip.CreateFloatingIP(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            self.floating_ip.floating_network_id,
        ]
        verifylist = [
            ('network', self.floating_ip.floating_network_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_ip.assert_called_once_with(
            **{
                'floating_network_id': self.floating_ip.floating_network_id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_all_options(self):
        arglist = [
            '--subnet',
            self.subnet.id,
            '--port',
            self.floating_ip.port_id,
            '--floating-ip-address',
            self.floating_ip.floating_ip_address,
            '--fixed-ip-address',
            self.floating_ip.fixed_ip_address,
            '--description',
            self.floating_ip.description,
            '--dns-domain',
            self.floating_ip.dns_domain,
            '--dns-name',
            self.floating_ip.dns_name,
            self.floating_ip.floating_network_id,
        ]
        verifylist = [
            ('subnet', self.subnet.id),
            ('port', self.floating_ip.port_id),
            ('fixed_ip_address', self.floating_ip.fixed_ip_address),
            ('network', self.floating_ip.floating_network_id),
            ('description', self.floating_ip.description),
            ('dns_domain', self.floating_ip.dns_domain),
            ('dns_name', self.floating_ip.dns_name),
            ('floating_ip_address', self.floating_ip.floating_ip_address),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_ip.assert_called_once_with(
            **{
                'subnet_id': self.subnet.id,
                'port_id': self.floating_ip.port_id,
                'floating_ip_address': self.floating_ip.floating_ip_address,
                'fixed_ip_address': self.floating_ip.fixed_ip_address,
                'floating_network_id': self.floating_ip.floating_network_id,
                'description': self.floating_ip.description,
                'dns_domain': self.floating_ip.dns_domain,
                'dns_name': self.floating_ip.dns_name,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_floating_ip_create_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
            self.floating_ip.floating_network_id,
        ]
        verifylist = [
            ('network', self.floating_ip.floating_network_id),
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_ip.assert_called_once_with(
            **{
                'floating_network_id': self.floating_ip.floating_network_id,
                'project_id': project.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_floating_ip_create_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        domain = identity_fakes_v3.FakeDomain.create_one_domain()
        self.projects_mock.get.return_value = project
        arglist = [
            "--project",
            project.name,
            "--project-domain",
            domain.name,
            self.floating_ip.floating_network_id,
        ]
        verifylist = [
            ('network', self.floating_ip.floating_network_id),
            ('project', project.name),
            ('project_domain', domain.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_ip.assert_called_once_with(
            **{
                'floating_network_id': self.floating_ip.floating_network_id,
                'project_id': project.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_floating_ip_with_qos(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy
        arglist = [
            '--qos-policy',
            qos_policy.id,
            self.floating_ip.floating_network_id,
        ]
        verifylist = [
            ('network', self.floating_ip.floating_network_id),
            ('qos_policy', qos_policy.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_ip.assert_called_once_with(
            **{
                'floating_network_id': self.floating_ip.floating_network_id,
                'qos_policy_id': qos_policy.id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def _test_create_with_tag(self, add_tags=True):
        arglist = [self.floating_ip.floating_network_id]
        if add_tags:
            arglist += ['--tag', 'red', '--tag', 'blue']
        else:
            arglist += ['--no-tag']

        verifylist = [
            ('network', self.floating_ip.floating_network_id),
        ]
        if add_tags:
            verifylist.append(('tags', ['red', 'blue']))
        else:
            verifylist.append(('no_tag', True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_ip.assert_called_once_with(
            **{
                'floating_network_id': self.floating_ip.floating_network_id,
            }
        )
        if add_tags:
            self.network_client.set_tags.assert_called_once_with(
                self.floating_ip, tests_utils.CompareBySet(['red', 'blue'])
            )
        else:
            self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_with_tags(self):
        self._test_create_with_tag(add_tags=True)

    def test_create_with_no_tag(self):
        self._test_create_with_tag(add_tags=False)


class TestDeleteFloatingIPNetwork(TestFloatingIPNetwork):
    # The floating ips to be deleted.
    floating_ips = network_fakes.FakeFloatingIP.create_floating_ips(count=2)

    def setUp(self):
        super().setUp()

        self.network_client.delete_ip.return_value = None

        # Get the command object to test
        self.cmd = fip.DeleteFloatingIP(self.app, None)

    def test_floating_ip_delete(self):
        self.network_client.find_ip.side_effect = [
            self.floating_ips[0],
            self.floating_ips[1],
        ]
        arglist = [
            self.floating_ips[0].id,
        ]
        verifylist = [
            ('floating_ip', [self.floating_ips[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.find_ip.assert_called_once_with(
            self.floating_ips[0].id,
            ignore_missing=False,
        )
        self.network_client.delete_ip.assert_called_once_with(
            self.floating_ips[0]
        )
        self.assertIsNone(result)

    def test_floating_ip_delete_multi(self):
        self.network_client.find_ip.side_effect = [
            self.floating_ips[0],
            self.floating_ips[1],
        ]
        arglist = []
        verifylist = []

        for f in self.floating_ips:
            arglist.append(f.id)
        verifylist = [
            ('floating_ip', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = [
            call(
                self.floating_ips[0].id,
                ignore_missing=False,
            ),
            call(
                self.floating_ips[1].id,
                ignore_missing=False,
            ),
        ]
        self.network_client.find_ip.assert_has_calls(calls)

        calls = []
        for f in self.floating_ips:
            calls.append(call(f))
        self.network_client.delete_ip.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_floating_ip_delete_multi_exception(self):
        self.network_client.find_ip.side_effect = [
            self.floating_ips[0],
            exceptions.CommandError,
        ]
        arglist = [
            self.floating_ips[0].id,
            'unexist_floating_ip',
        ]
        verifylist = [
            ('floating_ip', [self.floating_ips[0].id, 'unexist_floating_ip']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 floating IPs failed to delete.', str(e))

        self.network_client.find_ip.assert_any_call(
            self.floating_ips[0].id,
            ignore_missing=False,
        )
        self.network_client.find_ip.assert_any_call(
            'unexist_floating_ip',
            ignore_missing=False,
        )
        self.network_client.delete_ip.assert_called_once_with(
            self.floating_ips[0]
        )


class TestListFloatingIPNetwork(TestFloatingIPNetwork):
    # The floating ips to list up
    floating_ips = network_fakes.FakeFloatingIP.create_floating_ips(count=3)
    fake_network = network_fakes.create_one_network(
        {
            'id': 'fake_network_id',
        }
    )
    fake_port = network_fakes.create_one_port(
        {
            'id': 'fake_port_id',
        }
    )
    fake_router = network_fakes.create_one_router(
        {
            'id': 'fake_router_id',
        }
    )

    columns = (
        'ID',
        'Floating IP Address',
        'Fixed IP Address',
        'Port',
        'Floating Network',
        'Project',
    )
    columns_long = (
        *columns,
        'Router',
        'Status',
        'Description',
        'Tags',
        'DNS Name',
        'DNS Domain',
    )

    data = []
    data_long = []
    for ip in floating_ips:
        data.append(
            (
                ip.id,
                ip.floating_ip_address,
                ip.fixed_ip_address,
                ip.port_id,
                ip.floating_network_id,
                ip.project_id,
            )
        )
        data_long.append(
            (
                ip.id,
                ip.floating_ip_address,
                ip.fixed_ip_address,
                ip.port_id,
                ip.floating_network_id,
                ip.project_id,
                ip.router_id,
                ip.status,
                ip.description,
                ip.tags,
                ip.dns_domain,
                ip.dns_name,
            )
        )

    def setUp(self):
        super().setUp()

        self.network_client.ips.return_value = self.floating_ips
        self.network_client.find_network.return_value = self.fake_network

        self.network_client.find_port.return_value = self.fake_port
        self.network_client.find_router.return_value = self.fake_router

        # Get the command object to test
        self.cmd = fip.ListFloatingIP(self.app, None)

    def test_floating_ip_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_network(self):
        arglist = [
            '--network',
            'fake_network_id',
        ]
        verifylist = [
            ('networks', ['fake_network_id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'floating_network_id': ['fake_network_id'],
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_port(self):
        arglist = [
            '--port',
            'fake_port_id',
        ]
        verifylist = [
            ('ports', ['fake_port_id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'port_id': ['fake_port_id'],
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_fixed_ip_address(self):
        arglist = [
            '--fixed-ip-address',
            self.floating_ips[0].fixed_ip_address,
        ]
        verifylist = [
            ('fixed_ip_address', self.floating_ips[0].fixed_ip_address),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'fixed_ip_address': self.floating_ips[0].fixed_ip_address,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_floating_ip_address(self):
        arglist = [
            '--floating-ip-address',
            self.floating_ips[0].floating_ip_address,
        ]
        verifylist = [
            ('floating_ip_address', self.floating_ips[0].floating_ip_address),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'floating_ip_address': self.floating_ips[
                    0
                ].floating_ip_address,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with()
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_floating_ip_list_status(self):
        arglist = [
            '--status',
            'ACTIVE',
            '--long',
        ]
        verifylist = [
            ('status', 'ACTIVE'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'status': 'ACTIVE',
            }
        )
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_floating_ip_list_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.ips.assert_called_once_with(**filters)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
            '--project-domain',
            project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.ips.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_floating_ip_list_router(self):
        arglist = [
            '--router',
            'fake_router_id',
            '--long',
        ]
        verifylist = [
            ('routers', ['fake_router_id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'router_id': ['fake_router_id'],
            }
        )
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_list_with_tag_options(self):
        arglist = [
            '--tags',
            'red,blue',
            '--any-tags',
            'red,green',
            '--not-tags',
            'orange,yellow',
            '--not-any-tags',
            'black,white',
        ]
        verifylist = [
            ('tags', ['red', 'blue']),
            ('any_tags', ['red', 'green']),
            ('not_tags', ['orange', 'yellow']),
            ('not_any_tags', ['black', 'white']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.ips.assert_called_once_with(
            **{
                'tags': 'red,blue',
                'any_tags': 'red,green',
                'not_tags': 'orange,yellow',
                'not_any_tags': 'black,white',
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestShowFloatingIPNetwork(TestFloatingIPNetwork):
    def setUp(self):
        super().setUp()

        self.floating_ip = sdk_fakes.generate_fake_resource(
            _floating_ip.FloatingIP
        )
        self.network_client.find_ip.return_value = self.floating_ip

        self.columns = (
            'created_at',
            'description',
            'dns_domain',
            'dns_name',
            'fixed_ip_address',
            'floating_ip_address',
            'floating_network_id',
            'id',
            'name',
            'port_details',
            'port_id',
            'project_id',
            'qos_policy_id',
            'revision_number',
            'router_id',
            'status',
            'subnet_id',
            'tags',
            'updated_at',
        )
        self.data = (
            self.floating_ip.created_at,
            self.floating_ip.description,
            self.floating_ip.dns_domain,
            self.floating_ip.dns_name,
            self.floating_ip.fixed_ip_address,
            self.floating_ip.floating_ip_address,
            self.floating_ip.floating_network_id,
            self.floating_ip.id,
            self.floating_ip.name,
            format_columns.DictColumn(self.floating_ip.port_details),
            self.floating_ip.port_id,
            self.floating_ip.project_id,
            self.floating_ip.qos_policy_id,
            self.floating_ip.revision_number,
            self.floating_ip.router_id,
            self.floating_ip.status,
            self.floating_ip.subnet_id,
            self.floating_ip.tags,
            self.floating_ip.updated_at,
        )

        # Get the command object to test
        self.cmd = fip.ShowFloatingIP(self.app, None)

    def test_floating_ip_show(self):
        arglist = [
            self.floating_ip.id,
        ]
        verifylist = [
            ('floating_ip', self.floating_ip.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestSetFloatingIP(TestFloatingIPNetwork):
    # Fake data for option tests.
    floating_network = network_fakes.create_one_network()
    subnet = network_fakes.FakeSubnet.create_one_subnet()
    port = network_fakes.create_one_port()

    # The floating ip to be set.
    floating_ip = network_fakes.FakeFloatingIP.create_one_floating_ip(
        attrs={
            'floating_network_id': floating_network.id,
            'port_id': port.id,
            'tags': ['green', 'red'],
        }
    )

    def setUp(self):
        super().setUp()
        self.network_client.find_ip.return_value = self.floating_ip
        self.network_client.find_port.return_value = self.port
        self.network_client.update_ip.return_value = None
        self.network_client.set_tags.return_value = None

        # Get the command object to test
        self.cmd = fip.SetFloatingIP(self.app, None)

    def test_port_option(self):
        arglist = [
            self.floating_ip.id,
            '--port',
            self.floating_ip.port_id,
        ]
        verifylist = [
            ('floating_ip', self.floating_ip.id),
            ('port', self.floating_ip.port_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'port_id': self.floating_ip.port_id,
        }

        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )

        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def test_fixed_ip_option(self):
        arglist = [
            self.floating_ip.id,
            '--port',
            self.floating_ip.port_id,
            "--fixed-ip-address",
            self.floating_ip.fixed_ip_address,
        ]
        verifylist = [
            ('floating_ip', self.floating_ip.id),
            ('port', self.floating_ip.port_id),
            ('fixed_ip_address', self.floating_ip.fixed_ip_address),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'port_id': self.floating_ip.port_id,
            'fixed_ip_address': self.floating_ip.fixed_ip_address,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def test_description_option(self):
        arglist = [
            self.floating_ip.id,
            '--port',
            self.floating_ip.port_id,
            '--description',
            self.floating_ip.description,
        ]
        verifylist = [
            ('floating_ip', self.floating_ip.id),
            ('port', self.floating_ip.port_id),
            ('description', self.floating_ip.description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'port_id': self.floating_ip.port_id,
            'description': self.floating_ip.description,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def test_qos_policy_option(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy
        arglist = [
            "--qos-policy",
            qos_policy.id,
            self.floating_ip.id,
        ]
        verifylist = [
            ('qos_policy', qos_policy.id),
            ('floating_ip', self.floating_ip.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'qos_policy_id': qos_policy.id,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def test_port_and_qos_policy_option(self):
        qos_policy = network_fakes.create_one_qos_policy()
        self.network_client.find_qos_policy.return_value = qos_policy
        arglist = [
            "--qos-policy",
            qos_policy.id,
            '--port',
            self.floating_ip.port_id,
            self.floating_ip.id,
        ]
        verifylist = [
            ('qos_policy', qos_policy.id),
            ('port', self.floating_ip.port_id),
            ('floating_ip', self.floating_ip.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'qos_policy_id': qos_policy.id,
            'port_id': self.floating_ip.port_id,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def test_no_qos_policy_option(self):
        arglist = [
            "--no-qos-policy",
            self.floating_ip.id,
        ]
        verifylist = [
            ('no_qos_policy', True),
            ('floating_ip', self.floating_ip.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'qos_policy_id': None,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def test_port_and_no_qos_policy_option(self):
        arglist = [
            "--no-qos-policy",
            '--port',
            self.floating_ip.port_id,
            self.floating_ip.id,
        ]
        verifylist = [
            ('no_qos_policy', True),
            ('port', self.floating_ip.port_id),
            ('floating_ip', self.floating_ip.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        attrs = {
            'qos_policy_id': None,
            'port_id': self.floating_ip.port_id,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

    def _test_set_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['red', 'blue', 'green']
        else:
            arglist = ['--no-tag']
            verifylist = [('no_tag', True)]
            expected_args = []
        arglist.extend([self.floating_ip.id])
        verifylist.extend([('floating_ip', self.floating_ip.id)])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_ip.called)
        self.network_client.set_tags.assert_called_once_with(
            self.floating_ip, tests_utils.CompareBySet(expected_args)
        )
        self.assertIsNone(result)

    def test_set_with_tags(self):
        self._test_set_tags(with_tags=True)

    def test_set_with_no_tag(self):
        self._test_set_tags(with_tags=False)


class TestUnsetFloatingIP(TestFloatingIPNetwork):
    floating_network = network_fakes.create_one_network()
    subnet = network_fakes.FakeSubnet.create_one_subnet()
    port = network_fakes.create_one_port()

    # The floating ip to be unset.
    floating_ip = network_fakes.FakeFloatingIP.create_one_floating_ip(
        attrs={
            'floating_network_id': floating_network.id,
            'port_id': port.id,
            'tags': ['green', 'red'],
        }
    )

    def setUp(self):
        super().setUp()
        self.network_client.find_ip.return_value = self.floating_ip
        self.network_client.update_ip.return_value = None
        self.network_client.set_tags.return_value = None

        # Get the command object to test
        self.cmd = fip.UnsetFloatingIP(self.app, None)

    def test_floating_ip_unset_port(self):
        arglist = [
            self.floating_ip.id,
            "--port",
        ]
        verifylist = [
            ('floating_ip', self.floating_ip.id),
            ('port', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'port_id': None,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

        self.assertIsNone(result)

    def test_floating_ip_unset_qos_policy(self):
        arglist = [
            self.floating_ip.id,
            "--qos-policy",
        ]
        verifylist = [
            ('floating_ip', self.floating_ip.id),
            ('qos_policy', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'qos_policy_id': None,
        }
        self.network_client.find_ip.assert_called_once_with(
            self.floating_ip.id,
            ignore_missing=False,
        )
        self.network_client.update_ip.assert_called_once_with(
            self.floating_ip, **attrs
        )

        self.assertIsNone(result)

    def _test_unset_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['green']
        else:
            arglist = ['--all-tag']
            verifylist = [('all_tag', True)]
            expected_args = []
        arglist.append(self.floating_ip.id)
        verifylist.append(('floating_ip', self.floating_ip.id))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_ip.called)
        self.network_client.set_tags.assert_called_once_with(
            self.floating_ip, tests_utils.CompareBySet(expected_args)
        )
        self.assertIsNone(result)

    def test_unset_with_tags(self):
        self._test_unset_tags(with_tags=True)

    def test_unset_with_all_tag(self):
        self._test_unset_tags(with_tags=False)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import endpoint as _endpoint
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import region as _region
from openstack.identity.v3 import service as _service
from openstack.test import fakes as sdk_fakes

from openstackclient.identity.v3 import endpoint
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestEndpoint(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the EndpointManager Mock
        self.endpoints_mock = self.identity_client.endpoints
        self.endpoints_mock.reset_mock()
        self.ep_filter_mock = self.identity_client.endpoint_filter
        self.ep_filter_mock.reset_mock()

        # Get a shortcut to the ServiceManager Mock
        self.services_mock = self.identity_client.services
        self.services_mock.reset_mock()

        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains
        self.domains_mock.reset_mock()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()


class TestEndpointCreate(identity_fakes.TestIdentityv3):
    columns = (
        'enabled',
        'id',
        'interface',
        'region',
        'region_id',
        'service_id',
        'url',
        'service_name',
        'service_type',
    )

    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)
        self.endpoint = sdk_fakes.generate_fake_resource(
            resource_type=_endpoint.Endpoint,
            service_id=self.service.id,
            interface='admin',
            region_id=self.region.id,
        )

        self.identity_sdk_client.create_endpoint.return_value = self.endpoint
        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.get_region.return_value = self.region

        # Get the command object to test
        self.cmd = endpoint.CreateEndpoint(self.app, None)

    def test_endpoint_create_no_options(self):
        arglist = [
            self.service.id,
            self.endpoint.interface,
            self.endpoint.url,
        ]
        verifylist = [
            ('enabled', True),
            ('service', self.service.id),
            ('interface', self.endpoint.interface),
            ('url', self.endpoint.url),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Fake endpoints come with a region ID by default, so set it to None
        setattr(self.endpoint, "region_id", None)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'service_id': self.service.id,
            'url': self.endpoint.url,
            'interface': self.endpoint.interface,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_endpoint.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            True,
            self.endpoint.id,
            self.endpoint.interface,
            None,
            None,
            self.service.id,
            self.endpoint.url,
            self.service.name,
            self.service.type,
        )
        self.assertEqual(datalist, data)

    def test_endpoint_create_region(self):
        arglist = [
            self.service.id,
            self.endpoint.interface,
            self.endpoint.url,
            '--region',
            self.region.id,
        ]
        verifylist = [
            ('enabled', True),
            ('service', self.service.id),
            ('interface', self.endpoint.interface),
            ('url', self.endpoint.url),
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'service_id': self.service.id,
            'url': self.endpoint.url,
            'interface': self.endpoint.interface,
            'is_enabled': True,
            'region_id': self.region.id,
        }

        self.identity_sdk_client.create_endpoint.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            True,
            self.endpoint.id,
            self.endpoint.interface,
            self.region.id,
            self.region.id,
            self.service.id,
            self.endpoint.url,
            self.service.name,
            self.service.type,
        )
        self.assertEqual(datalist, data)

    def test_endpoint_create_enable(self):
        arglist = [
            self.service.id,
            self.endpoint.interface,
            self.endpoint.url,
            '--enable',
        ]
        verifylist = [
            ('enabled', True),
            ('service', self.service.id),
            ('interface', self.endpoint.interface),
            ('url', self.endpoint.url),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Fake endpoints come with a region ID by default, so set it to None
        setattr(self.endpoint, "region_id", None)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'service_id': self.service.id,
            'url': self.endpoint.url,
            'interface': self.endpoint.interface,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_endpoint.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            True,
            self.endpoint.id,
            self.endpoint.interface,
            None,
            None,
            self.service.id,
            self.endpoint.url,
            self.service.name,
            self.service.type,
        )
        self.assertEqual(datalist, data)

    def test_endpoint_create_disable(self):
        arglist = [
            self.service.id,
            self.endpoint.interface,
            self.endpoint.url,
            '--disable',
        ]
        verifylist = [
            ('enabled', False),
            ('service', self.service.id),
            ('interface', self.endpoint.interface),
            ('url', self.endpoint.url),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Fake endpoints come with a region ID by default, so set it to None
        setattr(self.endpoint, "region_id", None)
        setattr(self.endpoint, "is_enabled", False)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'service_id': self.service.id,
            'url': self.endpoint.url,
            'interface': self.endpoint.interface,
            'is_enabled': False,
        }

        self.identity_sdk_client.create_endpoint.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            False,
            self.endpoint.id,
            self.endpoint.interface,
            None,
            None,
            self.service.id,
            self.endpoint.url,
            self.service.name,
            self.service.type,
        )
        self.assertEqual(datalist, data)


class TestEndpointDelete(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.endpoint = sdk_fakes.generate_fake_resource(_endpoint.Endpoint)

        self.identity_sdk_client.find_endpoint.return_value = self.endpoint
        self.identity_sdk_client.delete_endpoint.return_value = None

        # Get the command object to test
        self.cmd = endpoint.DeleteEndpoint(self.app, None)

    def test_endpoint_delete(self):
        arglist = [
            self.endpoint.id,
        ]
        verifylist = [
            ('endpoint', [self.endpoint.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_endpoint.assert_called_with(
            self.endpoint.id,
        )
        self.assertIsNone(result)


class TestEndpointList(identity_fakes.TestIdentityv3):
    columns = (
        'ID',
        'Region',
        'Service Name',
        'Service Type',
        'Enabled',
        'Interface',
        'URL',
    )

    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)
        self.endpoint = sdk_fakes.generate_fake_resource(
            resource_type=_endpoint.Endpoint,
            service_id=self.service.id,
            interface='admin',
            region_id=self.region.id,
        )

        self.identity_sdk_client.endpoints.return_value = [self.endpoint]
        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.services.return_value = [self.service]
        self.identity_sdk_client.get_region.return_value = self.region

        # Get the command object to test
        self.cmd = endpoint.ListEndpoint(self.app, None)

    def test_endpoint_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.endpoints.assert_called_with()

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint.id,
                self.region.id,
                self.service.name,
                self.service.type,
                True,
                self.endpoint.interface,
                self.endpoint.url,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_endpoint_list_service(self):
        arglist = [
            '--service',
            self.service.id,
        ]
        verifylist = [
            ('service', self.service.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'service_id': self.service.id,
        }
        self.identity_sdk_client.endpoints.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint.id,
                self.region.id,
                self.service.name,
                self.service.type,
                True,
                self.endpoint.interface,
                self.endpoint.url,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_endpoint_list_interface(self):
        arglist = [
            '--interface',
            self.endpoint.interface,
        ]
        verifylist = [
            ('interface', self.endpoint.interface),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'interface': self.endpoint.interface,
        }
        self.identity_sdk_client.endpoints.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint.id,
                self.region.id,
                self.service.name,
                self.service.type,
                True,
                self.endpoint.interface,
                self.endpoint.url,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_endpoint_list_region(self):
        arglist = [
            '--region',
            self.region.id,
        ]
        verifylist = [
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'region_id': self.region.id,
        }
        self.identity_sdk_client.endpoints.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint.id,
                self.region.id,
                self.service.name,
                self.service.type,
                True,
                self.endpoint.interface,
                self.endpoint.url,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_endpoint_list_project_with_project_domain(self):
        project = sdk_fakes.generate_fake_resource(_project.Project)
        domain = sdk_fakes.generate_fake_resource(_domain.Domain)

        self.identity_sdk_client.project_endpoints.return_value = [
            self.endpoint
        ]
        self.identity_sdk_client.find_project.return_value = project

        arglist = ['--project', project.name, '--project-domain', domain.name]
        verifylist = [
            ('project', project.name),
            ('project_domain', domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.project_endpoints.assert_called_with(
            project=project.id
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint.id,
                self.region.id,
                self.service.name,
                self.service.type,
                True,
                self.endpoint.interface,
                self.endpoint.url,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestEndpointSet(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.endpoint = sdk_fakes.generate_fake_resource(
            resource_type=_endpoint.Endpoint,
            service_id=self.service.id,
            interface='admin',
        )

        self.identity_sdk_client.find_endpoint.return_value = self.endpoint
        self.identity_sdk_client.update_endpoint.return_value = self.endpoint
        self.identity_sdk_client.find_service.return_value = self.service

        # Get the command object to test
        self.cmd = endpoint.SetEndpoint(self.app, None)

    def test_endpoint_set_no_options(self):
        arglist = [
            self.endpoint.id,
        ]
        verifylist = [
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id
        )
        self.assertIsNone(result)

    def test_endpoint_set_interface(self):
        arglist = ['--interface', 'public', self.endpoint.id]
        verifylist = [
            ('interface', 'public'),
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'interface': 'public',
        }
        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_set_url(self):
        arglist = ['--url', 'http://localhost:5000', self.endpoint.id]
        verifylist = [
            ('url', 'http://localhost:5000'),
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'url': 'http://localhost:5000',
        }
        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_set_service(self):
        arglist = ['--service', self.service.id, self.endpoint.id]
        verifylist = [
            ('service', self.service.id),
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'service_id': self.service.id,
        }
        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_set_region(self):
        arglist = ['--region', 'e-rzzz', self.endpoint.id]
        verifylist = [
            ('region', 'e-rzzz'),
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'region_id': 'e-rzzz',
        }
        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_set_enable(self):
        arglist = ['--enable', self.endpoint.id]
        verifylist = [
            ('enabled', True),
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
        }
        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_set_disable(self):
        arglist = ['--disable', self.endpoint.id]
        verifylist = [
            ('disabled', True),
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': False,
        }
        self.identity_sdk_client.update_endpoint.assert_called_with(
            self.endpoint.id, **kwargs
        )
        self.assertIsNone(result)


class TestEndpointShow(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.region = sdk_fakes.generate_fake_resource(_region.Region)
        self.endpoint = sdk_fakes.generate_fake_resource(
            resource_type=_endpoint.Endpoint,
            service_id=self.service.id,
            interface='admin',
            region_id=self.region.id,
        )

        self.identity_sdk_client.find_endpoint.return_value = self.endpoint

        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.get_region.return_value = self.region

        # Get the command object to test
        self.cmd = endpoint.ShowEndpoint(self.app, None)

    def test_endpoint_show(self):
        arglist = [
            self.endpoint.id,
        ]
        verifylist = [
            ('endpoint', self.endpoint.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.find_endpoint.assert_called_with(
            self.endpoint.id, ignore_missing=False
        )

        collist = (
            'enabled',
            'id',
            'interface',
            'region',
            'region_id',
            'service_id',
            'url',
            'service_name',
            'service_type',
        )
        self.assertEqual(collist, columns)
        datalist = (
            True,
            self.endpoint.id,
            self.endpoint.interface,
            self.region.id,
            self.region.id,
            self.service.id,
            self.endpoint.url,
            self.service.name,
            self.service.type,
        )
        self.assertEqual(datalist, data)


class TestEndpointCreateServiceWithoutName(TestEndpointCreate):
    service = sdk_fakes.generate_fake_resource(
        resource_type=_service.Service,
        name='',
    )
    region = sdk_fakes.generate_fake_resource(_region.Region)
    endpoint = sdk_fakes.generate_fake_resource(
        resource_type=_endpoint.Endpoint,
        service_id=service.id,
        interface='admin',
        region_id=region.id,
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = endpoint.CreateEndpoint(self.app, None)


class TestEndpointListServiceWithoutName(TestEndpointList):
    service = sdk_fakes.generate_fake_resource(
        resource_type=_service.Service,
        name='',
    )
    region = sdk_fakes.generate_fake_resource(_region.Region)
    endpoint = sdk_fakes.generate_fake_resource(
        resource_type=_endpoint.Endpoint,
        service_id=service.id,
        interface='admin',
        region_id=region.id,
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = endpoint.ListEndpoint(self.app, None)


class TestEndpointShowServiceWithoutName(TestEndpointShow):
    service = sdk_fakes.generate_fake_resource(
        resource_type=_service.Service,
        name='',
    )
    region = sdk_fakes.generate_fake_resource(_region.Region)
    endpoint = sdk_fakes.generate_fake_resource(
        resource_type=_endpoint.Endpoint,
        service_id=service.id,
        interface='admin',
        region_id=region.id,
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = endpoint.ShowEndpoint(self.app, None)


class TestAddProjectToEndpoint(TestEndpoint):
    project = identity_fakes.FakeProject.create_one_project()
    domain = identity_fakes.FakeDomain.create_one_domain()
    service = identity_fakes.FakeService.create_one_service()
    endpoint = identity_fakes.FakeEndpoint.create_one_endpoint(
        attrs={'service_id': service.id}
    )

    new_ep_filter = identity_fakes.FakeEndpoint.create_one_endpoint_filter(
        attrs={'endpoint': endpoint.id, 'project': project.id}
    )

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.endpoints_mock.get.return_value = self.endpoint

        # Update the image_id in the MEMBER dict
        self.ep_filter_mock.create.return_value = self.new_ep_filter
        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain
        # Get the command object to test
        self.cmd = endpoint.AddProjectToEndpoint(self.app, None)

    def test_add_project_to_endpoint_no_option(self):
        arglist = [
            self.endpoint.id,
            self.project.id,
        ]
        verifylist = [
            ('endpoint', self.endpoint.id),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.ep_filter_mock.add_endpoint_to_project.assert_called_with(
            project=self.project.id, endpoint=self.endpoint.id
        )
        self.assertIsNone(result)

    def test_add_project_to_endpoint_with_option(self):
        arglist = [
            self.endpoint.id,
            self.project.id,
            '--project-domain',
            self.domain.id,
        ]
        verifylist = [
            ('endpoint', self.endpoint.id),
            ('project', self.project.id),
            ('project_domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.ep_filter_mock.add_endpoint_to_project.assert_called_with(
            project=self.project.id, endpoint=self.endpoint.id
        )
        self.assertIsNone(result)


class TestRemoveProjectEndpoint(TestEndpoint):
    project = identity_fakes.FakeProject.create_one_project()
    domain = identity_fakes.FakeDomain.create_one_domain()
    service = identity_fakes.FakeService.create_one_service()
    endpoint = identity_fakes.FakeEndpoint.create_one_endpoint(
        attrs={'service_id': service.id}
    )

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.endpoints_mock.get.return_value = self.endpoint

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain
        self.ep_filter_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = endpoint.RemoveProjectFromEndpoint(self.app, None)

    def test_remove_project_endpoint_no_options(self):
        arglist = [
            self.endpoint.id,
            self.project.id,
        ]
        verifylist = [
            ('endpoint', self.endpoint.id),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.ep_filter_mock.delete_endpoint_from_project.assert_called_with(
            project=self.project.id,
            endpoint=self.endpoint.id,
        )
        self.assertIsNone(result)

    def test_remove_project_endpoint_with_options(self):
        arglist = [
            self.endpoint.id,
            self.project.id,
            '--project-domain',
            self.domain.id,
        ]
        verifylist = [
            ('endpoint', self.endpoint.id),
            ('project', self.project.id),
            ('project_domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.ep_filter_mock.delete_endpoint_from_project.assert_called_with(
            project=self.project.id,
            endpoint=self.endpoint.id,
        )
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Security Group action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any, cast

from cliff import columns as cliff_columns
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common
from openstackclient.network import utils as network_utils

LOG = logging.getLogger(__name__)


def _format_network_security_group_rules(
    sg_rules: list[dict[str, Any]],
) -> str:
    # For readability and to align with formatting compute security group
    # rules, trim keys with caller known (e.g. security group and tenant ID)
    # or empty values.
    for sg_rule in sg_rules:
        empty_keys = [k for k, v in sg_rule.items() if not v]
        for key in empty_keys:
            sg_rule.pop(key)
        sg_rule.pop('security_group_id', None)
        sg_rule.pop('tenant_id', None)
        sg_rule.pop('project_id', None)
    return utils.format_list_of_dicts(sg_rules) or ""


def _format_compute_security_group_rule(sg_rule: dict[str, Any]) -> str:
    info = network_utils.transform_compute_security_group_rule(sg_rule)
    # Trim parent security group ID since caller has this information.
    info.pop('parent_group_id', None)
    # Trim keys with empty string values.
    keys_to_trim = [
        'ip_protocol',
        'ip_range',
        'port_range',
        'remote_security_group',
    ]
    for key in keys_to_trim:
        if key in info and not info[key]:
            info.pop(key)
    return utils.format_dict(info)


def _format_compute_security_group_rules(
    sg_rules: list[dict[str, Any]],
) -> str:
    rules = []
    for sg_rule in sg_rules:
        rules.append(_format_compute_security_group_rule(sg_rule))
    return utils.format_list(rules, separator='\n') or ""


class NetworkSecurityGroupRulesColumn(cliff_columns.FormattableColumn[Any]):
    def human_readable(self) -> str:
        return _format_network_security_group_rules(self._value)


class ComputeSecurityGroupRulesColumn(cliff_columns.FormattableColumn[Any]):
    def human_readable(self) -> str:
        return _format_compute_security_group_rules(self._value)


_formatters_network = {
    'security_group_rules': NetworkSecurityGroupRulesColumn,
}


_formatters_compute = {
    'rules': ComputeSecurityGroupRulesColumn,
}


def _get_columns(item: Any) -> tuple[tuple[str, ...], tuple[str, ...]]:
    # We still support Nova managed security groups, where we have tenant_id.
    column_map = {
        'security_group_rules': 'rules',
    }
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


# TODO(abhiraut): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class CreateSecurityGroup(command.ShowOne, common.NeutronCommandWithExtraArgs):
    _description = _("Create a new security group")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "name", metavar="<name>", help=_("New security group name")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Security group description"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Owner's project (name or ID)"),
        )
        stateful_group = parser.add_mutually_exclusive_group()
        stateful_group.add_argument(
            "--stateful",
            action='store_true',
            default=None,
            help=_("Security group is stateful (default)"),
        )
        stateful_group.add_argument(
            "--stateless",
            action='store_true',
            default=None,
            help=_("Security group is stateless"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        _tag.add_tag_option_to_parser_for_create(parser, _('security group'))
        return parser

    def _get_description(self, parsed_args: argparse.Namespace) -> str:
        if parsed_args.description is not None:
            return cast(str, parsed_args.description)
        else:
            return cast(str, parsed_args.name)

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        # Build the create attributes.
        attrs = {}
        attrs['name'] = parsed_args.name
        attrs['description'] = self._get_description(parsed_args)
        if parsed_args.stateful:
            attrs['stateful'] = True
        if parsed_args.stateless:
            attrs['stateful'] = False
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.identity
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            attrs['project_id'] = project_id
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        # Create the security group and display the results.
        obj = client.create_security_group(**attrs)
        # tags cannot be set when created, so tags need to be set later.
        _tag.update_tags_for_set(client, obj, parsed_args)
        display_columns, property_columns = _get_columns(obj)
        data = utils.get_item_properties(
            obj, property_columns, formatters=_formatters_network
        )
        return (display_columns, data)


class DeleteSecurityGroup(command.Command):
    _description = _("Delete security group(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            nargs="+",
            help=_("Security group(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for group in parsed_args.group:
            try:
                obj = client.find_security_group(group, ignore_missing=False)
                client.delete_security_group(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete security group with "
                        "name or ID '%(group)s': %(e)s"
                    ),
                    {'group': group, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.group)
            msg = _("%(result)s of %(total)s groups failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


# TODO(rauta): Use the SDK resource mapped attribute names once
# the OSC minimum requirements include SDK 1.0.
class ListSecurityGroup(command.Lister):
    _description = _("List security groups")
    FIELDS_TO_RETRIEVE = [
        'id',
        'name',
        'description',
        'project_id',
        'tags',
        'shared',
    ]

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only security groups with the specified project "
                "(name or ID)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)

        shared_group = parser.add_mutually_exclusive_group()
        shared_group.add_argument(
            '--share',
            action='store_true',
            dest='shared',
            default=None,
            help=_("List only security groups shared between projects"),
        )
        shared_group.add_argument(
            '--no-share',
            action='store_false',
            dest='shared',
            default=None,
            help=_("List only security groups not shared between projects"),
        )

        _tag.add_tag_filtering_option_to_parser(parser, _('security group'))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        filters = {}
        if parsed_args.project:
            identity_client = self.app.client_manager.identity
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            filters['project_id'] = project_id

        if parsed_args.shared is not None:
            filters['shared'] = parsed_args.shared

        _tag.get_tag_filtering_args(parsed_args, filters)
        data = client.security_groups(
            fields=self.FIELDS_TO_RETRIEVE, **filters
        )

        columns = (
            "id",
            "name",
            "description",
            "project_id",
            "tags",
            "is_shared",
        )
        column_headers = (
            "ID",
            "Name",
            "Description",
            "Project",
            "Tags",
            "Shared",
        )
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class SetSecurityGroup(common.NeutronCommandWithExtraArgs):
    _description = _("Set security group properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_("Security group to modify (name or ID)"),
        )
        parser.add_argument(
            '--name', metavar='<new-name>', help=_("New security group name")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("New security group description"),
        )
        stateful_group = parser.add_mutually_exclusive_group()
        stateful_group.add_argument(
            "--stateful",
            action='store_true',
            default=None,
            help=_("Security group is stateful (default)"),
        )
        stateful_group.add_argument(
            "--stateless",
            action='store_true',
            default=None,
            help=_("Security group is stateless"),
        )
        _tag.add_tag_option_to_parser_for_set(parser, _('security group'))
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_security_group(
            parsed_args.group, ignore_missing=False
        )
        attrs = {}
        if parsed_args.name is not None:
            attrs['name'] = parsed_args.name
        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description
        if parsed_args.stateful:
            attrs['stateful'] = True
        if parsed_args.stateless:
            attrs['stateful'] = False
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        # NOTE(rtheis): Previous behavior did not raise a CommandError
        # if there were no updates. Maintain this behavior and issue
        # the update.
        client.update_security_group(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_set(client, obj, parsed_args)


class ShowSecurityGroup(command.ShowOne):
    _description = _("Display security group details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_("Security group to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_security_group(
            parsed_args.group, ignore_missing=False
        )
        display_columns, property_columns = _get_columns(obj)
        data = utils.get_item_properties(
            obj, property_columns, formatters=_formatters_network
        )
        return (display_columns, data)


class UnsetSecurityGroup(command.Command):
    _description = _("Unset security group properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar="<group>",
            help=_("Security group to modify (name or ID)"),
        )
        _tag.add_tag_option_to_parser_for_unset(parser, _('security group'))
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_security_group(
            parsed_args.group, ignore_missing=False
        )

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_unset(client, obj, parsed_args)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Container v1 action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


class CreateContainer(command.Lister):
    _description = _("Create new container")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--public',
            action='store_true',
            default=False,
            help="Make the container publicly accessible",
        )
        parser.add_argument(
            '--storage-policy',
            help="Specify a particular storage policy to use.",
        )
        parser.add_argument(
            'containers',
            metavar='<container-name>',
            nargs="+",
            help=_('New container name(s)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        results = []
        for container in parsed_args.containers:
            if len(container) > 256:
                msg = _(
                    'Container name is %d characters long, the default limit'
                    ' is 256'
                )
                LOG.warning(msg, len(container))
            data = self.app.client_manager.object_store.container_create(
                container=container,
                public=parsed_args.public,
                storage_policy=parsed_args.storage_policy,
            )
            results.append(data)

        columns = ("account", "container", "x-trans-id")
        return (
            columns,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in results
            ),
        )


class DeleteContainer(command.Command):
    _description = _("Delete container")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--recursive',
            '-r',
            action='store_true',
            default=False,
            help=_('Recursively delete objects and container'),
        )
        parser.add_argument(
            'containers',
            metavar='<container>',
            nargs="+",
            help=_('Container(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        for container in parsed_args.containers:
            if parsed_args.recursive:
                objs = self.app.client_manager.object_store.object_list(
                    container=container
                )
                for obj in objs:
                    self.app.client_manager.object_store.object_delete(
                        container=container,
                        object=obj['name'],
                    )
            self.app.client_manager.object_store.container_delete(
                container=container,
            )


class ListContainer(command.Lister):
    _description = _("List containers")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--prefix",
            metavar="<prefix>",
            help=_("Filter list using <prefix>"),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        parser.add_argument(
            "--end-marker",
            metavar="<end-marker>",
            help=_("End anchor for paging"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_('List all containers (default is 10000)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        columns: tuple[str, ...] = ('Name',)
        if parsed_args.long:
            columns += ('Bytes', 'Count')

        kwargs = {}
        if parsed_args.prefix:
            kwargs['prefix'] = parsed_args.prefix
        if parsed_args.marker:
            kwargs['marker'] = parsed_args.marker
        if parsed_args.end_marker:
            kwargs['end_marker'] = parsed_args.end_marker
        if parsed_args.limit:
            kwargs['limit'] = parsed_args.limit
        if parsed_args.all:
            kwargs['full_listing'] = True

        data = self.app.client_manager.object_store.container_list(**kwargs)

        return (
            columns,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SaveContainer(command.Command):
    _description = _("Save container contents locally")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Container to save'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.app.client_manager.object_store.container_save(
            container=parsed_args.container,
        )


class SetContainer(command.Command):
    _description = _("Set container properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Container to modify'),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            required=True,
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property on this container "
                "(repeat option to set multiple properties)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.app.client_manager.object_store.container_set(
            parsed_args.container,
            properties=parsed_args.property,
        )


class ShowContainer(command.ShowOne):
    _description = _("Display container details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Container to display'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        data = self.app.client_manager.object_store.container_show(
            container=parsed_args.container,
        )
        if 'properties' in data:
            data['properties'] = format_columns.DictColumn(data['properties'])

        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class UnsetContainer(command.Command):
    _description = _("Unset container properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Container to modify'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            required=True,
            action='append',
            default=[],
            help=_(
                'Property to remove from container '
                '(repeat option to remove multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.app.client_manager.object_store.container_unset(
            parsed_args.container,
            properties=parsed_args.property,
        )
